/*

 Copyright (C) 1990-1995 Mark Adler, Richard B. Wales, Jean-loup Gailly,
 Kai Uwe Rommel and Igor Mandrichenko.
 Permission is granted to any individual or institution to use, copy, or
 redistribute this software so long as all of the original files are included,
 that it is not sold for profit, and that this copyright notice is retained.

*/

#include "zip.h"

#include <time.h>

#define MATCH shmatch


#define PAD 0
#define PATH_END '/'

#define MSDOS_DIR_ATTR 0x10

/* Library functions not in (most) header files */

#ifndef UTIL    /* the companion #endif is a bit of ways down ... */

int utime OF((char *, time_t *));

extern char *label;
local ulg label_time = 0;
local ulg label_mode = 0;

/* Local functions */
local char *readd OF((DIR *));



int procname(n)
char *n;                /* name to process */
/* Process a name or sh expression to operate on (or exclude).  Return
   an error code in the ZE_ class. */
{
  char *a;              /* path and name for recursion */
  DIR *d;               /* directory stream from opendir() */
  char *e;              /* pointer to name from readd() */
  int m;                /* matched flag */
  char *p;              /* path for recursion */
  struct stat s;        /* result of stat() */
  struct zlist far *z;  /* steps through zfiles list */

  if (strcmp(n, "-") == 0)   /* if compressing stdin */
    return newname(n, 0);
  else if (
#ifdef S_IFLNK          /* if symbolic links exist ... */
      linkput ? lstat(n, &s) :
#endif /* S_IFLNK */
      SSTAT(n, &s)
     )
  {
    /* Not a file or directory--search for shell expression in zip file */
    p = ex2in(n, 0, (int *)NULL);       /* shouldn't affect matching chars */
    m = 1;
    for (z = zfiles; z != NULL; z = z->nxt) {
      if (MATCH(p, z->zname))
      {
        z->mark = pcount ? filter(z->zname) : 1;
        if (verbose)
            fprintf(mesg, "zip diagnostic: %scluding %s\n",
               z->mark ? "in" : "ex", z->name);
        m = 0;
      }
    }
    free((zvoid *)p);
    return m ? ZE_MISS : ZE_OK;
  }

  /* Live name--use if file, recurse if directory */
  if ((s.st_mode & S_IFDIR) == 0)
  {
    /* add or remove name of file */
    if ((m = newname(n, 0)) != ZE_OK)
      return m;
  } else {
    /* Add trailing / to the directory name */
    if ((p = malloc(strlen(n)+2)) == NULL)
      return ZE_MEM;
    if (strcmp(n, ".") == 0) {
      *p = 0;  /* avoid "./" prefix and do not create zip entry */
    } else {
      strcpy(p, n);
      a = p + strlen(p);
      if (a[-1] != '/')
        strcpy(a, "/");
      if (dirnames && (m = newname(p, 1)) != ZE_OK) {
        free((zvoid *)p);
        return m;
      }
    }
    /* recurse into directory */
    if (recurse && (d = opendir(n)) != NULL)
    {
      while ((e = readd(d)) != NULL) {
        if (strcmp(e, ".") && strcmp(e, ".."))
        {
          if ((a = malloc(strlen(p) + strlen(e) + 1)) == NULL)
          {
            free((zvoid *)p);
            closedir(d);
            return ZE_MEM;
          }
          strcat(strcpy(a, p), e);
          if ((m = procname(a)) != ZE_OK)   /* recurse on name */
          {
            if (m == ZE_MISS)
              zipwarn("name not matched: ", a);
            else
              ziperr(m, a);
          }
          free((zvoid *)a);
        }
      }
      free((zvoid *)p);
      closedir(d);
    }
  } /* (s.st_mode & S_IFDIR) == 0) */
  return ZE_OK;
}

char *ex2in(x, isdir, pdosflag)
char *x;                /* external file name */
int isdir;              /* input: x is a directory */
int *pdosflag;          /* output: force MSDOS file attributes? */
/* Convert the external file name to a zip file name, returning the malloc'ed
   string or NULL if not enough memory. */
{
  char *n;              /* internal file name (malloc'ed) */
  char *t;              /* shortened name */
  int dosflag;
  char mem[10] = "";    /* member name */
  char ext[10] = "";     /* extension name */
  
  dosflag = dosify;  /* default for non-DOS non-OS/2 */

  /* Find starting point in name before doing malloc */
  for (t = x; *t == '/'; t++)
    ;

  /* Make changes, if any, to the copied name (leave original intact) */
  if (!pathput)
    t = last(t);

  /* Discard directory names with zip -rj */
  if (*t == '\0')
    return t;

  /* Malloc space for internal name and copy it */
  if ((n = malloc(strlen(t) + 1)) == NULL)
    return NULL;
  strcpy(n, t);
  if (t = strrchr(n, '(')) {
     *t = '\0';
     strcpy(mem,t+1);
     if (t = strchr(mem, ')')) *t = '\0';
     if (t = strrchr(n, '.')) t++;
     else t = n;
     strcpy(ext,t);
     strcpy(t,mem);
     strcat(t,".");
     strcat(t,ext);
  }
  if (t = strrchr(n, '.')) {
     while (--t > n)
        if (*t == '.')
          *t = '/';
  }
  n = strtoasc(n,t);

  if (isdir == 42) return n;      /* avoid warning on unused variable */

  if (dosify)
    msname(n);

  /* Returned malloc'ed name */
  if (pdosflag) 
    *pdosflag = dosflag;
  return n;
}


char *in2ex(n)
char *n;                /* internal file name */
/* Convert the zip file name to an external file name, returning the malloc'ed
   string or NULL if not enough memory. */
{
  char *x;              /* external file name */

  if ((x = malloc(strlen(n) + 1 + PAD)) == NULL)
    return NULL;
  strtoebc(x, n);
  return x;
}

/*
 * Should we use mktime all the time and remove invlocal ??? (OvdL)
 */
void stamp(f, d)
char *f;                /* name of file to change */
ulg d;                  /* dos-style time to change it to */
/* Set last updated and accessed time of file f to the DOS time d. */
{
  struct tm t;          /* argument for mktime() or invlocal() */
  time_t u[2];          /* argument for utime() */
  extern time_t mktime OF((struct tm *));

  /* Convert DOS time to time_t format in u[0] and u[1] */
  t.tm_sec = (int)(d << 1) & 0x3e;
  t.tm_min = (int)(d >> 5) & 0x3f;
  t.tm_hour = (int)(d >> 11) & 0x1f;
  t.tm_mday = (int)(d >> 16) & 0x1f;
  t.tm_mon = ((int)(d >> 21) & 0xf) - 1;
  t.tm_year = ((int)(d >> 25) & 0x7f) + 80;
  u[0] = u[1] = mktime(&t);

  utime(f, u);
}

ulg filetime(f, a, n)
char *f;                /* name of file to get info on */
ulg *a;                 /* return value: file attributes */
long *n;                /* return value: file size */
{
  FILE *stream;
  time_t ltime;

  if (strcmp(f, "-") == 0)   /* if compressing stdin */
    return 1;
  else {
     if ((stream = fopen(f, "r")) != (FILE *)NULL)
        {
        if (n != NULL) {
           *n = -1L;
/*         fseek(stream,0L,SEEK_END);
           *n = ftell(stream);  - don't work with "r" */
        }
        fclose(stream);

        time(&ltime);
        return unix2dostime(&ltime);/* was 1 */
        }
     else return ZE_OPEN;
  }
}

int set_extra_field(z)
struct zlist *z;
/* create extra field and change z->att if desired */
{
   fldata_t fdata;
   FILE *stream;
   char fname[65];
   char type[60];

/*translate_eol = 0;*/
  if (aflag == ASCII) {
     z->att = ASCII;
     return ZE_OK;
  }
  if (bflag)
    z->att = BINARY;
  else
    z->att = __EBCDIC;
  z->extra = (char *)malloc(sizeof(fdata)+2);
  if (z->extra == NULL) {
     printf("FLDATA : Unable to allocate memory !\n");
     return ZE_MEM;
  }
  z->cextra = (char *)malloc(sizeof(fdata)+2);
  if (z->cextra == NULL) {
     printf("FLDATA : Unable to allocate memory !\n");
     return ZE_MEM;
  }
  z->ext = (sizeof(fdata)+2);
  z->cext = (sizeof(fdata)+2);
  strtoebc(fname,z->zname);
  if (bflag)
    stream = fopen(fname,"rb,type=record");
  else
    stream = fopen(fname,"r");
  if (stream == NULL) {
     printf("FLDATA : Could not open file : %s !\n",fname);
     return ZE_NONE;
   }

  fldata(stream,fname,&fdata);
/*put the system ID */
#ifdef VM_CMS
  *(z->extra+1) = EF_VMCMS >> 8;
  *(z->extra) = EF_VMCMS & 0xFF;
  *(z->cextra+1) = EF_VMCMS >> 8;
  *(z->cextra) = EF_VMCMS & 0xFF;
#else
  *(z->extra+1) = EF_MVS >> 8;
  *(z->extra) = EF_MVS & 0xFF;
  *(z->cextra+1) = EF_MVS >> 8;
  *(z->cextra) = EF_MVS & 0xFF;
#endif
  *(z->extra+3) = sizeof(fdata) >> 8;
  *(z->extra+2) = sizeof(fdata) & 0xFF;
  *(z->cextra+3) = sizeof(fdata) >> 8;
  *(z->cextra+2) = sizeof(fdata) & 0xFF;

  memcpy(z->extra+4,&fdata,sizeof(fdata));
  memcpy(z->cextra+4,&fdata,sizeof(fdata));
  fclose(stream);
  return ZE_OK;
}

int deletedir(d)
char *d;                /* directory to delete */
/* Delete the directory *d if it is empty, do nothing otherwise.
   Return the result of rmdir(), delete(), or system().
   For VMS, d must be in format [x.y]z.dir;1  (not [x.y.z]).
 */
{
    return rmdir(d);
}

void version_local()
{
    printf("Compiled with %s under %s.\n", "C/370 2.1",
#ifdef VM_CMS
      "VM/CMS"
#else
      "MVS"
#endif
    );
}

#endif /* !UTIL */
